==> opytimizer-2.1.3-py3-none-any.whl: opytimizer/optimizers/evolutionary/iwo.py <==
"""Invasive Weed Optimization.
"""

import copy

from tqdm import tqdm

import opytimizer.math.random as r
import opytimizer.utils.constants as c
import opytimizer.utils.exception as ex
import opytimizer.utils.history as h
import opytimizer.utils.logging as l
from opytimizer.core.optimizer import Optimizer

logger = l.get_logger(__name__)


class IWO(Optimizer):
    """An IWO class, inherited from Optimizer.

    This is the designed class to define IWO-related
    variables and methods.

    References:
        A. R. Mehrabian and C. Lucas. A novel numerical optimization algorithm inspired from weed colonization.
        Ecological informatics (2006).

    """

    def __init__(self, algorithm='IWO', hyperparams=None):
        """Initialization method.

        Args:
            algorithm (str): Indicates the algorithm name.
            hyperparams (dict): Contains key-value parameters to the meta-heuristics.

        """

        # Override its parent class with the receiving hyperparams
        super(IWO, self).__init__(algorithm)

        # Minimum number of seeds
        self.min_seeds = 0

        # Maximum number of seeds
        self.max_seeds = 5

        # Exponent to calculate the Spatial Dispersal
        self.e = 2

        # Final standard deviation
        self.final_sigma = 0.001

        # Initial standard deviation
        self.init_sigma = 3

        # Now, we need to build this class up
        self._build(hyperparams)

        logger.info('Class overrided.')

    @property
    def min_seeds(self):
        """int: Minimum number of seeds.

        """

        return self._min_seeds

    @min_seeds.setter
    def min_seeds(self, min_seeds):
        if not isinstance(min_seeds, int):
            raise ex.TypeError('`min_seeds` should be an integer')
        if min_seeds < 0:
            raise ex.ValueError('`min_seeds` should be >= 0')

        self._min_seeds = min_seeds

    @property
    def max_seeds(self):
        """int: Maximum number of seeds.

        """

        return self._max_seeds

    @max_seeds.setter
    def max_seeds(self, max_seeds):
        if not isinstance(max_seeds, int):
            raise ex.TypeError('`max_seeds` should be an integer')
        if max_seeds < self.min_seeds:
            raise ex.ValueError('`max_seeds` should be >= `min_seeds`')

        self._max_seeds = max_seeds

    @property
    def e(self):
        """float: Exponent used to calculate the Spatial Dispersal.

        """

        return self._e

    @e.setter
    def e(self, e):
        if not isinstance(e, (float, int)):
            raise ex.TypeError('`e` should be a float or integer')
        if e < 0:
            raise ex.ValueError('`e` should be >= 0')

        self._e = e

    @property
    def final_sigma(self):
        """float: Final standard deviation.

        """

        return self._final_sigma

    @final_sigma.setter
    def final_sigma(self, final_sigma):
        if not isinstance(final_sigma, (float, int)):
            raise ex.TypeError('`final_sigma` should be a float or integer')
        if final_sigma < 0:
            raise ex.ValueError('`final_sigma` should be >= 0')

        self._final_sigma = final_sigma

    @property
    def init_sigma(self):
        """float: Initial standard deviation.

        """

        return self._init_sigma

    @init_sigma.setter
    def init_sigma(self, init_sigma):
        if not isinstance(init_sigma, (float, int)):
            raise ex.TypeError('`init_sigma` should be a float or integer')
        if init_sigma < 0:
            raise ex.ValueError('`init_sigma` should be >= 0')
        if init_sigma < self.final_sigma:
            raise ex.ValueError('`init_sigma` should be >= `final_sigma`')

        self._init_sigma = init_sigma

    def _spatial_dispersal(self, iteration, n_iterations):
        """Calculates the Spatial Dispersal coefficient (eq. 1).

        Args:
            iteration (int): Current iteration number.
            n_iterations (int): Maximum number of iterations.

        """

        # Calculating the iteration coefficient
        coef = ((n_iterations - iteration) ** self.e) / ((n_iterations + c.EPSILON) ** self.e)

        # Updating the Spatial Dispersial
        self.sigma = coef * (self.init_sigma - self.final_sigma) + self.final_sigma

    def _produce_offspring(self, agent, function):
        """Reproduces and flowers a seed into a new offpsring.

        Args:
            agent (Agent): An agent instance to be reproduced.
            function (Function): A Function object that will be used as the objective function.

        Returns:
            An evolved offspring.

        """

        # Makea a deepcopy on selected agent
        a = copy.deepcopy(agent)

        # For every possible decision variable
        for j, (lb, ub) in enumerate(zip(a.lb, a.ub)):
            # Updates its position
            a.position[j] += self.sigma * r.generate_uniform_random_number(lb, ub, a.n_dimensions)

        # Clips its limits
        a.clip_limits()

        # Calculates its fitness
        a.fit = function(a.position)

        return a

    def _update(self, agents, n_agents, function):
        """Method that wraps offsprings generations over all agents and variables.

        Args:
            agents (list): List of agents.
            n_agents (int): Number of possible agents in the space.
            function (Function): A Function object that will be used as the objective function.

        Returns:
            A new population with more fitted individuals.

        """

        # Creating a list for the produced offsprings
        offsprings = []

        # Sorting agents
        agents.sort(key=lambda x: x.fit)

        # Iterate through all agents
        for agent in agents:
            # Calculate the seeding ratio based on its fitness
            ratio = (agent.fit - agents[-1].fit) / (agents[0].fit - agents[-1].fit + c.EPSILON)

            # Calculates the number of produced seeds
            n_seeds = int(self.min_seeds + (self.max_seeds - self.min_seeds) * ratio)

            # For every seed
            for _ in range(n_seeds):
                # Reproduces and flowers the seed into a new agent
                a = self._produce_offspring(agent, function)

                # Appends the agent to the offsprings
                offsprings.append(a)

        # Joins both populations
        agents += offsprings

        # Performs a new sort on the merged population
        agents.sort(key=lambda x: x.fit)

        return agents[:n_agents]

    def run(self, space, function, store_best_only=False, pre_evaluation=None):
        """Runs the optimization pipeline.

        Args:
            space (Space): A Space object that will be evaluated.
            function (Function): A Function object that will be used as the objective function.
            store_best_only (bool): If True, only the best agent of each iteration is stored in History.
            pre_evaluation (callable): This function is executed before evaluating the function being optimized.

        Returns:
            A History object holding all agents' positions and fitness achieved during the task.

        """

        # Initial search space evaluation
        self._evaluate(space, function, hook=pre_evaluation)

        # We will define a History object for further dumping
        history = h.History(store_best_only)

        # Initializing a progress bar
        with tqdm(total=space.n_iterations) as b:
            # These are the number of iterations to converge
            for t in range(space.n_iterations):
                logger.file(f'Iteration {t+1}/{space.n_iterations}')

                # Calculates the current Spatial Dispersal
                self._spatial_dispersal(t, space.n_iterations)

                # Updating agents
                space.agents = self._update(space.agents, space.n_agents, function)

                # Checking if agents meet the bounds limits
                space.clip_limits()

                # After the update, we need to re-evaluate the search space
                self._evaluate(space, function, hook=pre_evaluation)

                # Every iteration, we need to dump agents and best agent
                history.dump(agents=space.agents, best_agent=space.best_agent)

                # Updates the `tqdm` status
                b.set_postfix(fitness=space.best_agent.fit)
                b.update()

                logger.file(f'Fitness: {space.best_agent.fit}')
                logger.file(f'Position: {space.best_agent.position}')

        return history

==> opytimizer-3.0.0-py3-none-any.whl: opytimizer/optimizers/evolutionary/iwo.py <==
"""Invasive Weed Optimization.
"""

import copy

import opytimizer.math.random as r
import opytimizer.utils.constant as c
import opytimizer.utils.exception as ex
import opytimizer.utils.logging as l
from opytimizer.core.optimizer import Optimizer

logger = l.get_logger(__name__)


class IWO(Optimizer):
    """An IWO class, inherited from Optimizer.

    This is the designed class to define IWO-related
    variables and methods.

    References:
        A. R. Mehrabian and C. Lucas. A novel numerical optimization algorithm inspired from weed colonization.
        Ecological informatics (2006).

    """

    def __init__(self, params=None):
        """Initialization method.

        Args:
            params (dict): Contains key-value parameters to the meta-heuristics.

        """

        # Overrides its parent class with the receiving params
        super(IWO, self).__init__()

        # Minimum number of seeds
        self.min_seeds = 0

        # Maximum number of seeds
        self.max_seeds = 5

        # Exponent to calculate the Spatial Dispersal
        self.e = 2

        # Final standard deviation
        self.final_sigma = 0.001

        # Initial standard deviation
        self.init_sigma = 3

        # Standard deviation
        self.sigma = 0

        # Builds the class
        self.build(params)

        logger.info('Class overrided.')

    @property
    def min_seeds(self):
        """int: Minimum number of seeds.

        """

        return self._min_seeds

    @min_seeds.setter
    def min_seeds(self, min_seeds):
        if not isinstance(min_seeds, int):
            raise ex.TypeError('`min_seeds` should be an integer')
        if min_seeds < 0:
            raise ex.ValueError('`min_seeds` should be >= 0')

        self._min_seeds = min_seeds

    @property
    def max_seeds(self):
        """int: Maximum number of seeds.

        """

        return self._max_seeds

    @max_seeds.setter
    def max_seeds(self, max_seeds):
        if not isinstance(max_seeds, int):
            raise ex.TypeError('`max_seeds` should be an integer')
        if max_seeds < self.min_seeds:
            raise ex.ValueError('`max_seeds` should be >= `min_seeds`')

        self._max_seeds = max_seeds

    @property
    def e(self):
        """float: Exponent used to calculate the Spatial Dispersal.

        """

        return self._e

    @e.setter
    def e(self, e):
        if not isinstance(e, (float, int)):
            raise ex.TypeError('`e` should be a float or integer')
        if e < 0:
            raise ex.ValueError('`e` should be >= 0')

        self._e = e

    @property
    def final_sigma(self):
        """float: Final standard deviation.

        """

        return self._final_sigma

    @final_sigma.setter
    def final_sigma(self, final_sigma):
        if not isinstance(final_sigma, (float, int)):
            raise ex.TypeError('`final_sigma` should be a float or integer')
        if final_sigma < 0:
            raise ex.ValueError('`final_sigma` should be >= 0')

        self._final_sigma = final_sigma

    @property
    def init_sigma(self):
        """float: Initial standard deviation.

        """

        return self._init_sigma

    @init_sigma.setter
    def init_sigma(self, init_sigma):
        if not isinstance(init_sigma, (float, int)):
            raise ex.TypeError('`init_sigma` should be a float or integer')
        if init_sigma < 0:
            raise ex.ValueError('`init_sigma` should be >= 0')
        if init_sigma < self.final_sigma:
            raise ex.ValueError('`init_sigma` should be >= `final_sigma`')

        self._init_sigma = init_sigma

    @property
    def sigma(self):
        """float: Standard deviation.

        """

        return self._sigma

    @sigma.setter
    def sigma(self, sigma):
        if not isinstance(sigma, (float, int)):
            raise ex.TypeError('`sigma` should be a float or integer')

        self._sigma = sigma

    def _spatial_dispersal(self, iteration, n_iterations):
        """Calculates the Spatial Dispersal coefficient (eq. 1).

        Args:
            iteration (int): Current iteration number.
            n_iterations (int): Maximum number of iterations.

        """

        # Calculates the iteration coefficient
        coef = ((n_iterations - iteration) ** self.e) / ((n_iterations + c.EPSILON) ** self.e)

        # Updates the Spatial Dispersial
        self.sigma = coef * (self.init_sigma - self.final_sigma) + self.final_sigma

    def _produce_offspring(self, agent, function):
        """Reproduces and flowers a seed into a new offpsring.

        Args:
            agent (Agent): An agent instance to be reproduced.
            function (Function): A Function object that will be used as the objective function.

        Returns:
            An evolved offspring.

        """

        # Makea a deepcopy on selected agent
        a = copy.deepcopy(agent)

        # For every possible decision variable
        for j, (lb, ub) in enumerate(zip(a.lb, a.ub)):
            # Updates its position
            a.position[j] += self.sigma * r.generate_uniform_random_number(lb, ub, a.n_dimensions)

        # Clips its limits
        a.clip_by_bound()

        # Calculates its fitness
        a.fit = function(a.position)

        return a

    def update(self, space, function, iteration, n_iterations):
        """Wraps Invasive Weed Optimization over all agents and variables.

        Args:
            space (Space): Space containing agents and update-related information.
            function (Function): A Function object that will be used as the objective function.
            iteration (int): Current iteration.
            n_iterations (int): Maximum number of iterations.

        """

        # Calculates the current Spatial Dispersal
        self._spatial_dispersal(iteration, n_iterations)

        # Calculates the number of agents and creates list of offsprings
        n_agents = len(space.agents)
        offsprings = []

        # Sorts agents
        space.agents.sort(key=lambda x: x.fit)

        # Iterates through all agents
        for agent in space.agents:
            # Calculates the seeding ratio based on its fitness
            ratio = (agent.fit - space.agents[-1].fit) / (space.agents[0].fit - space.agents[-1].fit + c.EPSILON)

            # Calculates the number of produced seeds
            n_seeds = int(self.min_seeds + (self.max_seeds - self.min_seeds) * ratio)

            # For every seed
            for _ in range(n_seeds):
                # Reproduces and flowers the seed into a new agent
                a = self._produce_offspring(agent, function)

                # Appends the agent to the offsprings
                offsprings.append(a)

        # Joins both populations, sorts and gathers best `n_agents`
        space.agents += offsprings
        space.agents.sort(key=lambda x: x.fit)
        space.agents = space.agents[:n_agents]
